html {
    background-color: white;
    font-family: -apple-system, BlinkMacSystemFont, "Segoe UI", Roboto, Oxygen-Sans, Ubuntu, Cantarell, "Helvetica Neue", sans-serif;;
    color: black;
    font-size: 90%;
}

h1 {
    text-align: left;
}

h2 {
    text-align: center;
    font-size: 20px;
}

.grid {
    display: grid;
    grid-template-columns: 1fr 1fr;
    grid-template-rows: auto 450px auto;
    row-gap: 15px;
    column-gap: 15px;
}

.parameters {
    display: flex;
    flex-direction: column;
    gap: 3px;
    /*justify-content: flex-end;*/
}

.parameter-single {
    display: flex;
    flex-direction: row;
    gap: 15px;
}

.parameter-single>.name {
    width: 30%;
}

.parameter-single>.value {
    width: 60%;
}

.parameter-single>.small-value {
    width: 15%;
}

.parameter-single>.space {
    width: 50%;
}

.parameter-double {
    display: flex;
    flex-direction: row;
    gap: 2px;
}

.parameter-double>.name {
    width: 40%;
}

.parameter-double>.value {
    width: 30%;
}

.small-parameters {
    display: flex;
    flex-direction: row;
    gap: 2px;
}

.small-parameters>.space {
    width: 40%;
}

.small-parameters>.name {
    width: 22%;
}

.small-parameters>.value {
    width: 8%;
}


.parameter-double-button {
    display: flex;
    flex-direction: row;
    gap: 10px;
}

.parameter-double-button>.name {
    width: 15%;
}

.parameter-double-button>.value {
    width: 10%;
}

.parameter-double-button>.button {
    width: 50%;
}

input { 
    text-align: right; 
}

.large-button {
    display: flex;
    flex-direction: row;
}

.large-buttons>.button1 {
    width: 50%;
}

.large-buttons>.button2 {
    width: 50%;
}

#log- {
    max-height: 150px;
    border: thin lightgrey solid;
    overflow-x: scroll;
    overflow-y: scroll;
}

/*
input:valid, input:invalid {
    outline: none black;
}
*/

[hidden] {
    display: none !important;
}

.large-buttons>button {
    font-size: small;
    display: inline-block;
    font-weight: bold;
    line-height: 0;
    padding: 1rem 1rem;
}

/*
label {
    font-weight: bold;
}
*/

details {
    padding: .8em;
    background: #eae9ff;
    border-radius: 5px;
    border: 2px solid #b1adff;
}

.plot-tools>.parameters {
    padding: .8em;
    background: #eae9ff;
    border-radius: 5px;
    border: 2px solid #b1adff;
}

details>summary {
    font-weight: bold;
    cursor: pointer;
}

/*
button {
    background-color: #eae9ff;
    border-radius: 5px;
    border: 2px solid #b1adff;
    text-align: center;
}
*/

.root {
    display: flex;
    flex-direction: column;
    gap: 15px;
}

pre {
    background-color: #ffffff;
}

.warnings {
    border: 2px solid orange;
    background: lightyellow;
}

.help {
    border: 2px solid buttonborder;
    background: buttonface;
}

details>.parameters {
    padding-top: .8em;
}

.graph {
    border-radius: 5px;
    border: 2px solid #b1adff;
    padding: .8em
}

/* no hide spin-button in numeric fields */
input::-webkit-outer-spin-button,
input::-webkit-inner-spin-button {
    -webkit-appearance: none;
    margin: 0;
}

input[type=number] {
    -moz-appearance:textfield;
}
